#!/usr/bin/perl
#  
#  compute-frequencies.pl, compute frequencies of words in a text
#  Copyright (C) 2001  Pascal Courtois, Benjamin Drieu
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

use Getopt::Std;

my %tabword, %taboccurs;
my $nbword;
my %opts;
my $text;
my $num = 39;

getopt('fhvnc', \%opts);


# Possibly set $num to user-specified value.
if (exists $opts{'n'})
{
    if (not $opts{'n'} =~ /^\d+$/)
    {
	print STDERR "compute-frequencies.pl: `$opts{'n'}' must be a positive integer.\n";
	exit 1;
    }
    $num = $opts{'n'} - 1;
}


# Print version information.
if (exists $opts{'v'})
{
    print <<EOF;
compute-frequencies.pl version 0.1

Copyright (C) 2001 Pascal Courtois and Benjamin Drieu.
This is free software; see the source for copying conditions.  There is NO
warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
EOF
    exit;
}


# Be nice, print an help message.
if (exists $opts{'h'})
{
    print <<EOF;
Usage: compute-frequencies.pl [options]

  -f FILE	Use FILE instead of STDIN.
  -n NUM	Print the NUMs highests frequencies (default 40).
  -c CODING	Use coding system instead of iso-8859-1
		Supported coding systems are iso-8859-1 (western
		Europe) and iso-8859-2 (eastern Europe)
  -v		Print version information.
  -h		Print this help message.
EOF
    exit;
}


# Get coding system (defaulted to iso-8859-1)
my %coding_systems = 
    (
     # Western Europe
     'iso-8859-1' =>
     'a-zA-Z\270\300-\326\300-\366\370-\377',
     
     # Eastern Europe
     'iso-8859-2' =>
     'a-zA-Z\241\243\245\246\251-\254\256\257\261\263\265\266\271-\274\276-\326\330-\366\370-\376'
     );

my $coding_system = $coding_systems{$opts{'c'} or 'iso-8859-1'};
if (not $coding_system)
{
    print STDERR "compute-frequencies.pl: Coding system `$opts{'c'}' not supported yet\n";
    exit 1;
}


# Open file or use stdin
if (exists $opts{'f'})
{
    open FILE, $opts{'f'} or die $!;
    $text = join "", <FILE>;
    close FILE;
}
else
{
    $text = join "", <STDIN>;
}


# Do nothing if text if empty.
exit 1 if length $text == 0;


# Count occurences of words in text.
foreach my $word ($text =~/([$coding_system]+?)(?:[\.,; ]|$)/gsi) 
{
    $word =~ tr/A-Z/a-z/;
    $tabword{$word} ++;
    $nbword ++;
}


# Compute frequencies of words.
foreach my $key (keys %tabword)
{
    $taboccurs{$key} = sprintf("%8.8f", $tabword{$key} / $nbword);
}


# Sort them and print highest frequencies.
my @sorted = sort {$taboccurs{$b} cmp $taboccurs{$a} 
		   or $a cmp $b} keys(%taboccurs); 

foreach my $key (@sorted[0 .. (($num<$#sorted)?$num:$#sorted)])
{
    print "(\"$key\" . $taboccurs{$key})\n";
}
